'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = _interopDefault(require('react'));

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

var ENDPOINTS_COUNT = 0;
var ENDPOINT_BITS = {};
var CACHE = {};

function getChangedBits(_ref, _ref2) {
  var prev = _ref.state;
  var next = _ref2.state;
  var mask = 0;

  for (var id in next) {
    if (prev[id] !== next[id]) {
      mask |= ENDPOINT_BITS[id];
    }
  }

  return mask;
}

var SapientContext = React.createContext({}, getChangedBits);
var PENDING = 0;
var RESOLVED = 1;
var REJECTED = 2;
var Sapient = function Sapient(_ref3) {
  var children = _ref3.children;

  var _React$useState = React.useState({}),
      _React$useState2 = _slicedToArray(_React$useState, 2),
      state = _React$useState2[0],
      setState = _React$useState2[1];

  return React.createElement(SapientContext.Provider, {
    value: {
      state: state,
      setState: setState
    }
  }, children);
};

function readCache(cache, fetch, id) {
  var result = cache[id];

  if (result === undefined) {
    var promise = fetch(id);
    cache[id] = {
      status: PENDING,
      value: promise
    };
    promise.then(function (value) {
      cache[id] = {
        status: RESOLVED,
        value: value
      };
    }, function (error) {
      cache[id] = {
        status: REJECTED,
        value: error
      };
    });
    throw promise;
  } else {
    var status = result.status,
        value = result.value;

    switch (status) {
      case PENDING:
        {
          throw value; // promise
        }

      case REJECTED:
        {
          throw value; // error
        }

      case RESOLVED:
        {
          return value; // cached value
        }

      default:
        {
          return null;
        }
    }
  }
}

function createEndpoint(name, methods) {
  var _endpoint;
  var endpointBits = ENDPOINT_BITS[name] = 1 << ENDPOINTS_COUNT++;
  var endpointCache = CACHE[name] = {};

  var invalidator = function invalidator() {
    var invalidatedValues = {};
    return [invalidatedValues, function (name, id) {
      invalidatedValues[name] = true;

      if (id === undefined) {
        CACHE[name] = {};
      } else {
        delete endpointCache[id];
      }
    }];
  };

  function triggerUpdate(context, invalidatedValues) {
    context.setState(function (state) {
      var nextState = Object.assign({}, state);
      Object.keys(invalidatedValues).forEach(function (name) {
        nextState[name] = (state[name] || 0) + 1;
      });
      return nextState;
    });
  }

  var creator = function creator(context) {
    return function (data) {
      var _invalidator = invalidator(),
          _invalidator2 = _slicedToArray(_invalidator, 2),
          invalidatedValues = _invalidator2[0],
          invalidate = _invalidator2[1];

      methods.create(data, invalidate).then(function (newId, newData) {
        var createdData = newData || data;
        endpointCache[newId] = {
          status: RESOLVED,
          value: createdData
        };
        triggerUpdate(context, invalidatedValues);
        return createdData;
      });
    };
  };

  var updater = function updater(context, id) {
    return function (data) {
      var _invalidator3 = invalidator(),
          _invalidator4 = _slicedToArray(_invalidator3, 2),
          invalidatedValues = _invalidator4[0],
          invalidate = _invalidator4[1];

      methods.update(id, data, invalidate).then(function (newData) {
        var updatedData = newData || data;
        invalidate(name, id);
        endpointCache[id] = {
          status: RESOLVED,
          value: updatedData
        };
        triggerUpdate(context, invalidatedValues);
        return updatedData;
      });
    };
  };

  var deleter = function deleter(context, id) {
    return function () {
      var _invalidator5 = invalidator(),
          _invalidator6 = _slicedToArray(_invalidator5, 2),
          invalidatedValues = _invalidator6[0],
          invalidate = _invalidator6[1];

      methods.delete(id, invalidate).then(function () {
        invalidate(name, id);
        triggerUpdate(context, invalidatedValues);
      });
    };
  };

  function accessEndpoint(context, id) {
    var _ref4;

    var value = readCache(endpointCache, methods.read, id);
    return [value, (_ref4 = {}, _defineProperty(_ref4, "create" + name, creator(context)), _defineProperty(_ref4, "update" + name, updater(context, id)), _defineProperty(_ref4, "delete" + name, deleter(context, id)), _ref4)];
  }

  var UseEndpoint =
  /*#__PURE__*/
  function (_React$Component) {
    _inherits(UseEndpoint, _React$Component);

    function UseEndpoint() {
      _classCallCheck(this, UseEndpoint);

      return _possibleConstructorReturn(this, _getPrototypeOf(UseEndpoint).apply(this, arguments));
    }

    _createClass(UseEndpoint, [{
      key: "render",
      value: function render() {
        var _this = this;

        return React.createElement(SapientContext.Consumer, {
          unstable_observedBits: endpointBits
        }, function (context) {
          var _this$props;

          return (_this$props = _this.props).children.apply(_this$props, _toConsumableArray(accessEndpoint(context, _this.props.id)));
        });
      }
    }]);

    return UseEndpoint;
  }(React.Component);

  var UseCreateEndpoint =
  /*#__PURE__*/
  function (_React$Component2) {
    _inherits(UseCreateEndpoint, _React$Component2);

    function UseCreateEndpoint() {
      _classCallCheck(this, UseCreateEndpoint);

      return _possibleConstructorReturn(this, _getPrototypeOf(UseCreateEndpoint).apply(this, arguments));
    }

    _createClass(UseCreateEndpoint, [{
      key: "render",
      value: function render() {
        var _this2 = this;

        return React.createElement(SapientContext.Consumer, {
          unstable_observedBits: endpointBits
        }, function (context) {
          return _this2.props.children(creator(context));
        });
      }
    }]);

    return UseCreateEndpoint;
  }(React.Component);

  var endpoint = (_endpoint = {}, _defineProperty(_endpoint, "Use" + name, UseEndpoint), _defineProperty(_endpoint, "UseCreate" + name, UseCreateEndpoint), _endpoint);

  if (React.useContext) {
    var useHook = function useHook(id) {
      var context = React.useContext(SapientContext, endpointBits);
      return accessEndpoint(context, id);
    };

    var useCreateHook = function useCreateHook() {
      var context = React.useContext(SapientContext, endpointBits);
      return creator(context);
    };

    endpoint["use" + name] = useHook;
    endpoint["useCreate" + name] = useCreateHook;
  }

  return endpoint;
}

function createApi(endpoints) {
  var api = {};

  for (var name in endpoints) {
    Object.assign(api, createEndpoint(name, endpoints[name]));
  }

  return api;
}

exports.SapientContext = SapientContext;
exports.Sapient = Sapient;
exports.createApi = createApi;
